###########################################################################
#
#         NConv
#
###########################################################################


NConv <- function(X, B=1000, alpha=0.05, Emp=TRUE, Conv=TRUE, CI=TRUE,
                  pHat=FALSE, t.zero=1e-15, t.P=1e-8, verbose=FALSE,
                  m.zero=1e-7,max.iter=100) {
  # Estimation and confidence interval of the number of species in a population
  # INPUTS
  # X : abundance distribution.
  #     Vector of integers, with length larger or equal to 2.
  #     X[i] number of species seen i times
  # B = 1000 : number of simulations for estimating the asymptotic distribution
  #     of the test statistic.
  # alpha = 0.05 : the asymptotic probability that the number of species lies
  #         outside the confidence interval.  0 < alpha < 0.5
  # Emp = TRUE for estimating the number of species with the
  #            empirical estimator of the abundance distribution.
  # Conv = TRUE for estimating the number of species using the
  #        least squares estimator of the abundance distribution under the
  #        constraint of convexity.
  # CI = TRUE for calculating the confidence interval
  # pHat = TRUE to save the estimated distribution under the
  #             constraint of convexity
  #
  # inputs for the function pConv. This function estimates the
  # abundance distribution under the constraint of convexity. 
  #    t.zero = 1e-15: non-zero minimum value
  #    t.P = 1e-8: threshold for testing that pHat  sums to one
  #    verbose = FALSE: if TRUE, print for each iteration
  #              on the  maximum support : pi, Psi and sumP
  #            pi : values of the positive measure pi at the last iteration
  #            Psi : value of the criterion to be minimised
  #            sumP : sum(pHat) at convergence
  #
  # inputs for the function tiedConv. This function calculates the
  # projection of a vector on the space of piecewise convex vectors.
  #    m.zero = 1e-7 : non-zero minimum value
  #    maw.iter = 100 : maximum number of iterations
  #
  # OUTPUT: a list
  # Nhat : The estimated values of the total number of species
  # CI : Lower and upper bounds of the confidence intervals for the
  #  total number of species
  # pHat : if Conv = TRUE and pHat=TRUE, estimated values of the truncated
  #    abundance distribution, and of its knots
  #
  if (!is.vector(X)) stop("The argument should be a vector")
  if (length(X)<2)
    stop("The argument should be a vector of length larger than 2")
  if (min(X) < 0) stop("The values of X should be non-negative")
  if ((alpha<0)|(alpha>=0.5)) stop("alpha should be between 0 and 0.5")
  alphaTest=c(1-alpha/2,alpha/2)
  if (sum(X)==0)
    stop("The total number of observed species should be strictly positive")
  f <- X/sum(X)
  D=sum(X)
  Nhat <- NULL
  ConfInt <- NULL
  if (Emp) {
    NEmp <- 2*X[1]-X[2]+D
    Nhat <-NEmp
    names(Nhat) <- 'Emp'
    if (CI) {
      if (X[1]==0)
        message("The number of species seen 1 time equals 0. The estimated standard-error of the empirical estimator equals 0")
      se <-  sqrt(6*X[1])
      ConfInt <-matrix(round(NEmp-se*qnorm(alphaTest)),nrow=1,ncol=2)
      dimnames(ConfInt) <- list(c('Emp'),c('lb','ub'))
    }
  }
  if (Conv) {
    pH <- pConv(f, t.zero=t.zero, t.P=t.P, verbose=verbose)
    if (pH$cvge != 0) {
      message("The algorithm for estimating the  abundance distribution under the constraint of convexity does not converge")
      print(paste("Error message:",pH$trace))
    }
    if (pH$cvge == 0) {
      if (pHat) {
        pp <- as.vector(pH$p)
        names(pp) <- paste("l=",1:length(pp),sep="")
        pCvx <- list(p=pp,knots=sort((pH$Spi+1)))
      }
      tau=length(pH$p)
      tetaChap <- 2*pH$p[1]-pH$p[2]+1
      NChap <- D*tetaChap
      Nhat <- c(Nhat,round(NChap))
      if (Emp) names(Nhat) <- c('Emp','CvxPi')
      if (!Emp) names(Nhat) <- 'CvxPi'
      if (CI) {
        noeuds <- sort((pH$Spi+1))
        NbNds <- length(noeuds)
        Nds <- noeuds
        if (NbNds > 1) {
          DeltaNds <- noeuds[-1] - noeuds[-NbNds]
          wh <- which(DeltaNds==1)
          if (length(wh)>0) Nds <- noeuds[1:wh[1]]
        }
        k=Nds[length(Nds)]
        if (k==2) {
          Cvge <- NaN
          qSPivot <- sqrt(6*pH$p[1])*qnorm(p=alphaTest)
          ICPlugin <-  c(NChap-sqrt(D)*qSPivot[1],NChap-sqrt(D)*qSPivot[2])
        }
        if (k>2) {
          if (k==tau+1) {
            V <- -pH$p[1:tau]%*%t(pH$p[1:tau])
            diag(V) <- pHat$p[1:tau]*(1-pH$p[1:tau])
          }
          if (k<=tau) {
            V <- -pH$p[1:k]%*%t(pH$p[1:k])
            diag(V) <- pH$p[1:k]*(1-pH$p[1:k])
          }
          W <- mvrnorm(n=B,mu=rep(0,ncol(V)),Sigma=V)
          if (k==tau+1) W <- cbind(W,rep(0,B))
          PhiW <- c(NaN,B)
          CvgeB <- c(NaN,B)
          for (i in 1:B) {
            rr <- tiedConv(W[i,],Nds,max.iter=max.iter,zero=m.zero)
            PhiW[i] <- 2*rr$fitted[1]-rr$fitted[2]
            CvgeB[i] <- rr$cvge
          }
          Cvge <- sum(CvgeB==0)
          if (Cvge/B < 1) {
            message(paste("The algorithm for calculating the quantiles of the asymptotic distribution failed to converge in",B-Cvge,"simulations over",B))
            if (Cvge > 0.5) {
              message("You may try to change the parameters m.zero or max.iter")
            }
            if (Cvge ==0 ) {
              message("The confidence interval based on the asymptotic distribution of the convex estimator cannot be calculated")
            }
          }
          if (Cvge>0) {
            SPivot <- PhiW[CvgeB==0]+rnorm(Cvge)*sqrt(tetaChap*(tetaChap-1))
            qSPivot <- quantile(SPivot, p=alphaTest, na.rm=TRUE)
            ICPlugin <-
              c(NChap-sqrt(D)*qSPivot[1],NChap-sqrt(D)*qSPivot[2])
          }
        } # fin if (k>2) {
        if (!Emp) {
          ConfInt <-matrix(round(ICPlugin),nrow=1,ncol=2)
          dimnames(ConfInt) <- list(c('CvxPi'),c('lb','ub'))
        }
        if (Emp) {
          ConfInt <- rbind(ConfInt,round(ICPlugin))
          dimnames(ConfInt)[[1]] <- c('Emp','CvxPi')
        }
      } # fin if (CI) {
    } # fin if (pH$cvge == 0) {
  } # fin if (Conv) {
  if ((!is.null(Nhat))&(is.null(ConfInt))) r=Nhat
  if ((!is.null(Nhat))&(!is.null(ConfInt)))
                  r=list(NHat=Nhat,CI=ConfInt)
  if ((Conv)&(pHat)&(pH$cvge==0)) r$pHat = pCvx
  return(r)
}


###########################################################################
#
#         pConv
#
###########################################################################

pConv <- function(ptild,  t.zero=1e-15, t.P=1e-8, max.sn=NULL, verbose=FALSE){
  # Calculation of pHat with support 0:LHat
  #
  # INPUT
  # ptild : vector of empirical probabilities
  #         with M+1 components
  #         where M=max(Support(ptild))
  # t.zero : non-zero minimum value
  # t.P : threshold for testing that pHat  sums to one
  # max.sn : maximum of the support of pHat
  #          By default max.sn=10*M
  # verbose : if TRUE, print for each iteration
  #           on the  maximum support : pi, Psi and sumP (see OUTPUT below)
  #
  # OUTPUT
  # res : list with components
  #   cvge = 0 if the criterion Psi decreases with the support of pi
  #   cvge = 1 if Psi increases
  #   cvge = 2 if maximum number of iterations reached
  #   code
  #   Spi : Support of the positive measure pi at the last iteration
  #   pi : values of the positive measure pi at the last iteration
  #   p : values of pHat
  #   Psi : scaler value of the criterion to be minimised
  #   sumP : scalar = sum(pHat) at convergence
  # history : data frame with components
  #   L : the maximum of the support of pi
  #   Psi : value of the criterion for the value L
  #   SumP : value of sum(pHat)
  M <- length(ptild)-1
  if (is.null(max.sn)) max.sn=10*M
  L1=M+1
  Psi <- NULL
  SumP <- NULL
  code <- NULL
  i=0
  valL <- round(seq(L1,max.sn,length=20))
  suiteL <- NULL
  for (L in valL) {
    i=i+1
    suiteL <- c(suiteL,L)
    if (i>=2) res1=resL
    resL <- Estim(L, ptild, t.zero)
    if (verbose==TRUE) {
      print(paste("maximum support =",L))
      print(resL$pi[order(resL$Spi)])
      print(paste("Crit Psi=",resL$Psi,"Sum Proba=",resL$SumP))
    }
    if (resL$code>0) {
      code=c(code,resL$code)
      Psi <- c(Psi,resL$Psi)
      SumP <- c(SumP,resL$SumP )
      if (L==max.sn) {
        trace=paste("The maximum number of iterations, max.sn=",
          max.sn,
          " is reached. The convergence is not reached with Code=",
          resL$code)
        history <- data.frame(L=suiteL,code=code,Psi=Psi,SumP=SumP)
        pi <- resL$pi
        names(pi) <- paste("j=",resL$Spi,sep="")
        pi <- pi[order(resL$Spi)]
        Spi <- sort(resL$Spi)
        res <- list(cvge=2,trace=trace,history=history,
                    Spi=Spi,pi=pi,p=resL$p,
                    Psi=resL$Psi,SumP=resL$SumP)
        return(res)
      }
    }
    if (resL$code ==0) {
      code=c(code,resL$code)
      Psi <- c(Psi,resL$Psi)
      SumP <- c(SumP,resL$SumP)
      if ((i==1)&(abs(resL$SumP-1)<t.P)) {
        history <- data.frame(L=L1,code=code,Psi=Psi,SumP=SumP)
        trace="Convergence is reached"
        res <- list(cvge=0,trace=trace,history=history,
                    Spi=resL$Spi,pi=resL$pi,p=resL$p,
                    Psi=resL$Psi,SumP=resL$SumP)
        return(res)
      }
      if (i>1) {
        if (abs(resL$SumP-1)<t.P) {
          history <- data.frame(L=suiteL,code=code,Psi=Psi,SumP=SumP)
          trace="Convergence is reached"
          pi <- resL$pi
          names(pi) <- paste("j=",resL$Spi,sep="")
          pi <- pi[order(resL$Spi)]
          Spi <- sort(resL$Spi)
          res <- list(cvge=0,trace=trace,history=history,
                      Spi=Spi,pi=pi,p=resL$p,
                      Psi=resL$Psi,SumP=resL$SumP)
          return(res)
        }
        DeltaPsi <-Psi[length(Psi)]-Psi[length(Psi)-1]
        if (DeltaPsi > 0) {
          trace=paste("The algoritm stopped because for L=",L,
            ",the criterion increases. Psi(",suiteL[i-1],")=",
            Psi[length(Psi)-1],"Psi(",suiteL[i],")=",Psi[length(Psi)])
          history <- data.frame(L=suiteL,code=code,Psi=Psi,SumP=SumP)
          pi <- res1$pi
          names(pi) <- paste("j=",res1$Spi,sep="")
          pi <- pi[order(res1$Spi)]
          Spi <- sort(res1$Spi)
          res <- list(cvge=1,trace=trace,history=history,
                      Spi=Spi,pi=pi,p=res1$p,
                      Psi=res1$Psi,SumP=res1$SumP)
          return(res)
        }
      }
    }
  }
}

###########################################################################
#
#         tiedConv
#
###########################################################################

tiedConv <- function(Y,bp,max.iter=100,zero=1e-7) {
  # Projection on piecewise convex functions
  # INPUTS
  # Y vector with n components
  # bp vector with r components : the break points
  # max.iter
  # zero
  # OUTPUT
  n <- length(Y)
  r  <-  length(bp)+1
  int <- matrix(NA,nrow=2,ncol=r)
  int[1,] <- c(1,bp)
  int[2,]<- c(bp,n)
  u0 <- Y
  end=1
  SCR <- NULL
  # iteration 1
  Is <- matrix(0,nrow=n,ncol=r)
  incr <- rep(0,n)
  it=0
#  u=Y
  while ((end >0)&(it<=max.iter)) {
    it=it+1
    Ism1=Is
    Is <- 0*Ism1
    Incr.s=0
#    usm1=u
    for (j in 1:r) {
      if (j>2) Incr.s  <-  apply(Is[,1:(j-1)],1,sum)
      if (j==2) Incr.s  <-  Is[,1]
      if (j < (r-1)) Incr.sm1 <- apply(Ism1[,(j+1):r],1,sum)
      if (j == (r-1)) Incr.sm1 <- Ism1[,r]
      if (j == r) Incr.sm1 <- 0
      u <- Y + Incr.s + Incr.sm1
      pts.j <- int[1,j]:int[2,j]
      res.cvx <- conreg(pts.j,u[pts.j],convex=TRUE,verbose=FALSE)
      u[pts.j] <- res.cvx$yf
      Is[,j] <- u-(Y + Incr.s + Incr.sm1)
    }
    SCR <- c(SCR, sum((Y-u)**2))
#    sum((usm1-u)**2)
    if (it >= 2) {
      if (abs(SCR[it] - SCR[it-1])<zero) end=0
    }
  }
#  print(c(it,max.iter))
#   if (it>=(max.iter+1)) message("Nbre max d'iterations atteint")
  return(list(fitted=u,nb.iter=it,cvge=end,SCR=SCR))
}


Step4 <- function(ptild, Sk, pik, Addj, T0) {
  #  updates the support Sk if necessary
  #
  # INPUT
  # ptild : vector of empirical probabilities
  # Sk : vector. Support of pi.
  # pik : vector with the same length as Spi. Values for the positive
  #       measure
  # Addj : point added to the support at Step2
  # T0 : matrix of j-th triangular functions for
  # j varying from 1 to L
  #
  # OUTPUT
  # code : scalar
  #  code = 0. The support Sk and the values pik has been successfully
  #            updated
  #
  #  code =1. It was not possible to find a new support.
  #
  # If code=0,
  #  NewS : the new support
  #  Newpi : values of the positive measure
  #
  # If code=1,
  #  Trace : nature of the problem
  # 
  cond=0
  Sk1=Sk
  pik1=pik
  M <- length(ptild)
  while (cond==0) {
    m <- max(c(Sk1, Addj))
    np <- length(c(Sk1, Addj))
    if (m >= M) {
      emp <- rep(0,m)
      emp[1:M] <- ptild
    # projection on the new support
      lesT <-t(T0[c(Sk1, Addj),1:m])
    }
    if (m < M) {
      emp <- ptild
      lesT <-t(T0[c(Sk1, Addj),1:M])
    }  
    if (np==1) reg <- lm(emp~t(lesT)-1)
    if (np>1) reg <- lm(emp~lesT-1)
    if (is.na(reg$coeff[np] < 0)){ 
      Sk1 <- NULL            
      pik1 <- NULL           
    }
    if ( !is.na(reg$coeff[np] < 0) ) {
      if (reg$coeff[np] < 0) {
        cond=1
        trace=paste("Step2 added",Addj,
          "Step 4 : its coefficient is negative:",reg$coeff[indNeg])
        return(list(code=1,trace=trace))
      }
    # if not
      if (reg$coeff[np] >= 0) {
        indNeg <- reg$coeff[1:(np-1)]<0
      # if the coefficients corresponding to Sk1 are all positive
      # then return to Estim to continue the algorithm
        if (sum(indNeg) ==  0) {
          cond=1
          return(list(code=0,NewS=c(Sk1,Addj),Newpi= reg$coeff))
        }
      # if some coefficients corresponding to Sk1 are negative
      # then suppress one of them
        if (sum(indNeg) >= 1) {
          if (np==2) {
            Sk1 <- NULL
            pik1 <- NULL
          }
          if (np>2) {
            amoinsb  <- pik1-reg$coeff[-np]
            ii <- which(amoinsb>0)
            l <- which.min(pik1[ii]/amoinsb[ii])
            Sk1 <- Sk1[-ii[l]]
            pik1 <- pik1[-ii[l]]
          }
        }
      # and calculate the projection on this new support
      } #  if (reg$coeff[np] >= 0) {
    } # if ( !is.na(reg$coeff[np] < 0) ) {
  } # while (cond==0) {
}


Step2 <- function(emp,Spi,pi,T0,t.zero){
  # calculates the directional derivatives D
  # and add a point to the support of pi if necessary
  #
  # INPUT
  # emp : vector of dimension L, whose M+1 first coordinates equal ptild
  # Spi : vector. Support of pi. max(Spi)=L
  # pi : vector with the same length as Spi. Values for the positive measure
  # The L-th trinagular function
  #
  # OUTPUT
  # D vector with L components. Directional derivatives of Psi
  # Addj : scalar. Addj = min(D). If  Addj <0, the point Addj must be
  # added to the support of pi
#  m <- max(Spi)
  m=dim(T0)[1]
  D <- rep(0,m)
  # calculation of the directional derivatives
  # D[j], for j varying from 1 to L
  for(j in 1:m){
    d <- rep(0,j) 
    for(l in 1:j){
      d[l] <- T0[j,l]*(sum(pi*T0[Spi,l])-emp[l])
    }
    D[j] <- sum(d)                        
  }
  # mise a 0 des valeurs inferieures au seuil
  if(sum(abs(D) < t.zero) != 0 ){ D[abs(D) < t.zero] <- 0}
  # If there exists at least one D[j] negative, then
  # add to the support Spi the value of j that minimizes D.
  minD <- min(D)
  if (minD >=0) Addj=0
  if (minD <0) Addj <- which.min(D)
  return( list(Addj=Addj, D=D) )
}
#

Tri1 <- function(J){
  # calculation of the matrix of j-th triangular functions for
  # j varying from 1 to J
  #
  # INPUT
  # J scalar
  #
  # OUTPUT
  # t : matrix with J rows and J columns
  # t(j,l) = 2*(j-(l-1))/(j*(j+1)) for 1<=l<=j
  t <- matrix(0,J,J)
  for(j in 1:J){
    t[j,1:j] <- 2*(j-(0:(j-1)))/(j*(j+1))
  }
  dimnames(t) <- list(paste("j=",1:J,sep=""),
                   paste("l=",0:(J-1),sep=""))
  return(t)
}

Estim <- function(L, ptild, t.zero){
  # calculation of pHat when the max(Supp(pi))=L
  #
  # INPUT
  # L : scalar : max of the support of pi
  # ptild : vector of empirical probabilities
  # t.zero : non-zero minimum value
  #
  # OUTPUT
  # code : integer
  #        code=0 : the convergence is reached
  #        code=1 : the convergence is not reached
  #                 (no positive measure is found)
  #  Spi : vector of integers.  Support of pi
  #  pi : vector with the same length as Spi.
  #       Values of the positive measure
  #  p : vector with L components. Values of the convex function p
  #  SumP : scalar. Sum of the components of p
  #  Psi : scalar. value of the crierion Psi
  #  trace : if code=1, nature of the problem
  emp <- rep(0,L)
  M <- length(ptild)
  emp[1:M] <- ptild
  iter <- 0
  # Initial value for (pi, Spi) : the L-th triangular function
  #
  # calculation of the matrix of j-th triangular functions for
  # j varying from 1 to L
  T0 <- Tri1(L)
  pi0 <-sum(emp*T0[L,])/sum(T0[L,]**2)
  Spi0 <- c(L)
  p0 <- pi0*T0[L,]
  Psi0 <- sum(p0^2)/2-sum((emp[1:length(p0)])*p0)
  SumP0 <- sum(p0)
  # The function Step2 calculates the directional derivatives D
  # and add a point to the support of pi if necessary
  St2 <- Step2(emp,Spi0,pi0,T0,t.zero)
  #
  # If St2$Addj==0 then the convergence is reached
  # when the max(Supp(pi))=L
  if (St2$Addj==0) {
   return(list(code=0, Spi=Spi0, pi=pi0, p=p0, Psi=Psi0, SumP=SumP0))
  }
  #
  # If St2$Addj !=0
  # Add one point to the support, and update the support in order to
  # get a positive measure
  Sk = Spi0
  pik = pi0
  pk=p0
  code=1
  while(code==1 ) {
    # the function Step4 updates the support Sk if necessary
    St4 <- Step4(ptild, Sk, pik, St2$Addj, T0)
     # if St4$code=1, then return to the main function for continuing
    # with a new value of L
    if (St4$code >=1 ) {
      return(list(code=St4$code, trace=St4$trace,
                  Spi=Spi0, pi=pi0, p=p0, Psi=Psi0, SumP=SumP0))
    }
    # if not, then go to step 2 with the new positive measure
    if (St4$code==0) {
      Sk=St4$NewS
      pik=St4$Newpi
      St2 <- Step2(emp,Sk,pik,T0,t.zero)
      # if St2$Addj=0, all the derivative are positive and the minimum
      # of Psi is reached (on the support with maximum smaller than L)
      if (St2$Addj==0) {
        code=0
        p = matrix(pik,nrow=1,ncol=length(pik))%*%T0[Sk,1:max(Sk)]
        Psi <- sum(p^2)/2-sum((emp[1:length(p)])*p)
        SumP <- sum(p)
        return(list(code=code, Spi=Sk, pi=pik, p=p, Psi=Psi, SumP=SumP))
      }
    }
  }
}


Emp <- function(X) {   
  # compute the empirical frequence of X
  supp <- seq(1,max(X),by=1)
  count <- rep(0,length(supp))
  for(i in 1:length(supp)) {
    count[i] <- sum(X==supp[i])
  }
  return(list(supp=supp,freq=count))
}

plot.NConv <- function(X,r) {
  nX <- length(X)
  p <- c(r$pHat$p,0)
  np <- length(p)
  rr <- range(c(X/sum(X),p))
  reps <- max(rr)/50
  plot(c(1,max(nX,np)),rr,type="n",
       xlab="",ylab="")
  title(main="Truncated abundance distribution")
  points(1:nX,X/sum(X),pch=1,lwd=2)
  points(1:np,p,pch=4,col=2,lwd=2)
  k <- r$pHat$knots
  m <- 
  segments(x0=k,y0=rep(-0.01,length(k)),x1=k,y1=p[r$pHat$knots],col=2)
  points(k,rep(-reps,length(k)),pch=3,col=2)
  legend(max(nX,np)/2,max(rr)-reps,legend=c("Empirical","Convex","knots"),
         pch=c(1,4,3),col=c(1,2,2),cex=1.5)
}
